#!/usr/bin/env python
# coding: utf-8

# In[4]:


import tkinter as tk
from tkinter import colorchooser, ttk
from tkinter.filedialog import askopenfilename, asksaveasfilename
from PIL import Image, ImageTk

# Layout constants.
MARGIN = 5
WINDOW_WID = 400
WINDOW_HGT = 300

class ScribbleSaveRestoreApp:
    # Create and manage the tkinter interface.
    def __init__(self):
        # Make the main interface.
        self.window = tk.Tk()
        self.window.title('scribble_save_restore')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry(f'{WINDOW_WID}x{WINDOW_HGT}')

        # Build menus.
        self.build_menus()

        # Build the toolbar.
        self.canvas = None
        self.build_toolbar()

        # Build the rest of the UI here...
        self.canvas = tk.Canvas(self.window, bg='white',
            borderwidth=2, relief=tk.SUNKEN)
        self.canvas.pack(padx=10, pady=(0, 10),
            side=tk.BOTTOM, fill=tk.BOTH, expand=True)

        # Initially we are not drawing anything.
        self.line_points = None
        self.current_line = None

        self.start_point = None
        self.end_point = None
        self.current_oval = None

        # Initially select the arrow.
        self.select_tool('tool_arrow')

        # Display the window.
        self.window.focus_force()
        self.window.mainloop()

    def build_menus(self):
        menu_bar = tk.Menu(self.window)
        self.window.config(menu=menu_bar)

        # File menu
        self.build_file_menu(menu_bar)

    def build_file_menu(self, menu_bar):
        '''Build the File menu.'''
        file_menu = tk.Menu(menu_bar, tearoff=False)
        menu_bar.add_cascade(label='File', menu=file_menu)

        # Note: You must save the photo image in a persistent
        # variable so tkinter can redraw it when needed.

        # Note: You need to include the compound parameter
        # if you want to use both images and text. That value
        # tells where to put the image relative to the text.

        # New
        self.new_photo_image = ImageTk.PhotoImage(file='new.png')        
        file_menu.add_command(label='New', accelerator='Ctrl+N', compound='left',
            command=self.mnu_new, image=self.new_photo_image)
        self.window.bind('<Control-n>', lambda event: self.mnu_new())

        # Open
        self.open_photo_image = ImageTk.PhotoImage(file='open.png')
        file_menu.add_command(label='Open...', accelerator='Ctrl+O', compound='left',
            command=self.mnu_open, image=self.open_photo_image)
        self.window.bind('<Control-o>', lambda event: self.mnu_open())

        file_menu.add_separator()

        # Note: If you don't include an image, the menu item doesn't
        # quite line up with items that do have an image. You can fix
        # it by including a blank image.

        # Save As
        self.save_as_photo_image = ImageTk.PhotoImage(file='save_as.png')
        file_menu.add_command(label='Save As...', accelerator='Ctrl+S', compound='left',
            command=self.mnu_save_as, image=self.save_as_photo_image)
        self.window.bind('<Control-s>', lambda event: self.mnu_save_as())

        file_menu.add_separator()

        # Exit
        self.exit_photo_image = ImageTk.PhotoImage(file='exit.png')
        file_menu.add_command(label='Exit', compound='left',
            command=self.mnu_exit, image=self.exit_photo_image)

    def mnu_new(self):
        self.canvas.delete(tk.ALL)
    def mnu_open(self):
        filetypes = [
            ('JSON Files', '*.json'),
            ('All Files', '*,*')]
        filename = askopenfilename(filetypes=filetypes, defaultextension='.json')
        if len(filename) == 0: return
        with open(filename, 'r') as file:
            json_data = file.read()
        deserialize_canvas(self.canvas, json_data)

    def mnu_save_as(self):
        filetypes = [
            ('JSON Files', '*.json'),
            ('All Files', '*,*')]
        filename = asksaveasfilename(filetypes=filetypes, defaultextension='.json')
        if len(filename) == 0: return
        json_data = serialize_canvas(self.canvas)
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(json_data)
        print('Saved')

    def mnu_exit(self):
        self.kill_callback()

    def build_toolbar(self):
        '''Build the toolbar.'''
        # Create a toolbar
        self.toolbar = tk.Frame(self.window, relief=tk.RIDGE, borderwidth=2)

        # Add buttons and other tools to the toolbar.
        self.tool_padx = self.tool_pady = 3

        # Color button.
        self.fg_color = 'black'
        self.make_color_button(self.fg_color)

        # Line thickness OptionMenu.
        self.make_thickness_option_menu()

        # Line dash OptionMenu.
        self.make_dash_option_menu()

        # Tools.
        self.make_tools()

        # Place the toolbar
        self.toolbar.pack(side=tk.TOP, fill=tk.X, padx=5, pady=5)

    def make_color_button(self, color):
        self.tool_hgt = 16
        self.tool_wid = 16
        self.fg_button_image = tk.PhotoImage(width=self.tool_wid, height=self.tool_hgt)
        y1 = int(self.tool_hgt / 2)
        y2 = y1 + 2
        self.fg_button_image.put(color, to=(0,y1,self.tool_wid-1,y2))
        self.fg_button = tk.Button(self.toolbar, command=self.choose_color, image=self.fg_button_image)
        self.fg_button.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)

        # Start with black selected.
        self.fg_color = 'black'

    def make_thickness_option_menu(self):
        # Make a dictionary mapping thickness values to images.
        # Note: When I tried using strings holding the numbers 1 through 5
        # or integers holding the numbers 1 through 5, the menu seemed to get
        # confused so the first menu item showed the text '1'.
        # Hence you need to add 1 to the selected option to get the
        # desired line thickness.
        self.thickness_images = {
            0: tk.PhotoImage(file='thickness_1.png'),
            1: tk.PhotoImage(file='thickness_2.png'),
            2: tk.PhotoImage(file='thickness_3.png'),
            3: tk.PhotoImage(file='thickness_4.png'),
            4: tk.PhotoImage(file='thickness_5.png'),
        }

        # Make the OptionMenu.
        self.thickness_var = tk.IntVar(value=0)
        self.thickness_om = tk.OptionMenu(self.toolbar,
            self.thickness_var, 0, 1, 2, 3, 4,
            command=self.thickness_selected)
        self.thickness_om.config(indicatoron=False)
        thickness = self.thickness_var.get()
        self.thickness_om.configure(image=self.thickness_images[thickness])

        # Get the underlying menu.
        menu = self.thickness_om['menu']

        # Add the menu item images.
        for number, image in self.thickness_images.items():
            menu.entryconfigure(number, image=image, compound=tk.NONE)

        self.thickness_om.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)

    def make_dash_option_menu(self):
        # Not all dash patterns produce results on all platforms. See:
        # https://stackoverflow.com/questions/41796792/tkinter-canvas-dash-option-is-not-behaving-as-expected
        # Make a dictionary mapping dash values to images.
        self.dash_images = {
            'dash_solid':       tk.PhotoImage(file='dash_solid.png'),
            'dash_6_4':         tk.PhotoImage(file='dash_6_4.png'),
            'dash_6_4_2_4':     tk.PhotoImage(file='dash_6_4_2_4.png'),
            'dash_6_4_2_4_2_4': tk.PhotoImage(file='dash_6_4_2_4_2_4.png'),
        }

        # Make the OptionMenu.
        self.dash_var = tk.StringVar(value='dash_solid')
        self.dash_om = tk.OptionMenu(self.toolbar, self.dash_var,
            'dash_solid', 'dash_6_4', 'dash_6_4_2_4', 'dash_6_4_2_4_2_4',
            command=self.dash_selected)
        self.dash_om.config(indicatoron=False)
        self.dash_om.configure(image=self.dash_images['dash_solid'])

        # Get the underlying menu.
        menu = self.dash_om['menu']

        # Add the menu item images.
        for name, image in self.dash_images.items():
            menu.entryconfigure(name, image=image, compound=tk.NONE)

        self.dash_om.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)

        # Start with solid lines.
        self.dash_pattern = ()

    def make_tools(self):
        separator = ttk.Separator(self.toolbar, orient=tk.VERTICAL)
        separator.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady, fill=tk.Y)

        self.tool_images = {
            'tool_arrow':    tk.PhotoImage(file='tool_arrow.png'),
            'tool_scribble': tk.PhotoImage(file='tool_scribble.png'),
            'tool_oval':     tk.PhotoImage(file='tool_oval.png'),
        }
        self.tool_buttons = {}
        for name, image in self.tool_images.items():
            button = tk.Button(self.toolbar, image=image,
                command=lambda bound_name=name: self.select_tool(bound_name))
            button.pack(side=tk.LEFT, padx=self.tool_padx, pady=self.tool_pady)
            self.tool_buttons[name] = button

    def mnu_exit(self):
        self.kill_callback()

    def kill_callback(self):
        self.window.destroy()

    def choose_color(self):
        color = colorchooser.askcolor(title='Foreground Color',
            initialcolor=self.fg_color)
        if color[1] is not None:
            # Save the selected color.
            self.fg_color = color[1]

            # Fill the color button with the selected color.
            y1 = int(self.tool_hgt / 2)
            y2 = y1 + 2
            self.fg_button_image.put(color[1], to=(0,y1,self.tool_wid-1,y2))

    def thickness_selected(self, option_num):
        # Give the OptionMenu the correct image.
        self.thickness_om.configure(image=self.thickness_images[option_num])

    def dash_selected(self, option_name):
        # Save the selected dash pattern.
        match self.dash_var.get():
            case 'dash_solid':
                self.dash_pattern = ()
            case 'dash_6_4':
                self.dash_pattern = (6, 4)
            case 'dash_6_4_2_4':
                self.dash_pattern = (6, 4, 2, 4)
            case 'dash_6_4_2_4_2_4':
                self.dash_pattern = (6, 4, 2, 4, 2, 4)

        # Give the OptionMenu the correct image.
        self.dash_om.configure(image=self.dash_images[option_name])

    def scribble_mouse_down(self, event):
        # Save the starting point.
        self.line_points = [(event.x, event.y)]

    def scribble_mouse_move(self, event):
        # Ignore extraneous events.
        if self.line_points is None: return

        # If there is a current line, remove it.
        if self.current_line is not None:
            self.canvas.delete(self.current_line)

        # Create a line.
        self.line_points.append((event.x, event.y))
        thickness, dash_pattern, fg_color = self.get_drawing_parameters()
        self.current_line = self.canvas.create_line(
            self.line_points, fill=fg_color, width=thickness,
            dash=dash_pattern)

    def scribble_mouse_up(self, event):
        # Ignore extraneous events.
        if self.line_points is None: return

        # We are no longer creating a line.
        self.line_points = None
        self.current_line = None

    def oval_mouse_down(self, event):
        # Save the starting point.
        self.start_point = (event.x, event.y)

    def oval_mouse_move(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # If there is a current oval, remove it.
        if self.current_oval is not None:
            self.canvas.delete(self.current_oval)

        # Create an oval.
        self.end_point = (event.x, event.y)
        thickness, dash_pattern, color = self.get_drawing_parameters()
        self.current_oval = self.canvas.create_oval(
            self.start_point, self.end_point, outline=self.fg_color,
            width=thickness, dash=dash_pattern)

    def oval_mouse_up(self, event):
        # Ignore extraneous events.
        if self.start_point is None: return

        # We are no longer creating an oval.
        self.start_point = None
        self.end_point = None
        self.current_oval = None

    def get_drawing_parameters(self):
        thickness = self.thickness_var.get() + 1
        return thickness, self.dash_pattern, self.fg_color

    def select_tool(self, tool_name):
        # Save the selected tool.
        self.selected_tool = tool_name

        # Unbind the mouse events.
        self.canvas.unbind('<Button-1>')
        self.canvas.unbind('<B1-Motion>')
        self.canvas.unbind('<ButtonRelease-1>')

        # Push the selected button down.
        for name, button in self.tool_buttons.items():
            if name == tool_name:
                button.config(relief=tk.SUNKEN)
            else:
                button.config(relief=tk.RAISED)

        # Set the appropriate cursor.
        match tool_name:
            case 'tool_arrow':
                self.canvas.config(cursor='')
            case 'tool_scribble':
                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.scribble_mouse_down)
                self.canvas.bind('<B1-Motion>', self.scribble_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.scribble_mouse_up)
            case 'tool_oval':
                self.canvas.config(cursor='cross')
                self.canvas.bind('<Button-1>', self.oval_mouse_down)
                self.canvas.bind('<B1-Motion>', self.oval_mouse_move)
                self.canvas.bind('<ButtonRelease-1>', self.oval_mouse_up)


# In[7]:


import tkinter as tk
import json

def serialize_canvas(canvas):
    items = []
    for item in canvas.find_all():
        item_data = {
            'type':    canvas.type(item),
            'coords':  canvas.coords(item),
            'options': {}
        }
        for option in canvas.itemconfig(item):
            value = canvas.itemcget(item, option)
            if value != '':
                item_data['options'][option] = value
        items.append(item_data)

    # Return the serialized data.
    return json.dumps(items, indent=4)

def deserialize_canvas(canvas, data):
    # Delete any existing items.
    canvas.delete(tk.ALL)

    # Deserialize the JSON data.
    items = json.loads(data)

    # Recreate the items.
    for item in items:
        coords = item['coords']
        options = item['options']
        if item['type'] == 'line':
            canvas.create_line(*coords, **options)
        elif item['type'] == 'oval':
            canvas.create_oval(*coords, **options)


# In[8]:


ScribbleSaveRestoreApp()
print('Done')


# In[ ]:




